// 
// Includes
// 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "SS_Tiny85.h"

//
// Globals
//

uint16_t _rx_delay_centering = 0;
uint16_t _rx_delay_intrabit = 0;
uint16_t _rx_delay_stopbit = 0;
uint16_t _tx_delay = 0;

uint16_t _buffer_overflow = false;

// static data
static char _receive_buffer[_SS_MAX_RX_BUFF];
static volatile uint8_t _receive_buffer_tail;
static volatile uint8_t _receive_buffer_head;

// private methods
#ifndef TX_ONLY
  #define rx_pin_read() (SERPIN & (1<<RXPIN))
#endif

// private static method for timing
static inline void tunedDelay(uint16_t delay);

//
// Lookup table
//
typedef struct _DELAY_TABLE {
	long baud;
	unsigned short rx_delay_centering;
	unsigned short rx_delay_intrabit;
	unsigned short rx_delay_stopbit;
	unsigned short tx_delay;
} DELAY_TABLE;

#if F_CPU == 1000000

// At 1Mhz anything over 4800 is so error ridden it will not work
static const PROGMEM DELAY_TABLE table[] =
{
  //  baud    rxcenter    rxintra    rxstop  tx
  { 4800,     14,        28,       27,    27,    },
  { 2400,     28,        56,       56,    56,    }
};

const int XMIT_START_ADJUSTMENT = 3;

#elif F_CPU == 8000000

static const PROGMEM DELAY_TABLE table[] =
{
	//  baud    rxcenter    rxintra    rxstop  tx
  { 19200,    20,         55,        55,     52,     },
	{ 9600,     50,         114,       114,    112,    }
};

const int XMIT_START_ADJUSTMENT = 4;

#elif F_CPU == 16000000

static const PROGMEM DELAY_TABLE table[] =
{
	//  baud    rxcenter   rxintra    rxstop    tx
  { 19200,    54,        117,       117,      114,   },
	{ 9600,     114,       236,       236,      233,   }
};

const int XMIT_START_ADJUSTMENT = 5;

#endif

//
// Private methods
//

/* static */
inline void tunedDelay(uint16_t delay) {
	uint8_t tmp = 0;

	asm volatile("sbiw    %0, 0x01 \n\t"
			"ldi %1, 0xFF \n\t"
			"cpi %A0, 0xFF \n\t"
			"cpc %B0, %1 \n\t"
			"brne .-10 \n\t"
			: "+w" (delay), "+a" (tmp)
			: "0" (delay)
	);
}

#ifndef TX_ONLY
  //
  // Interrupt handling, receive routine
  //
  ISR(PCINT0_vect) {
  	uint8_t d = 0;
  
  	// If RX line is high, then we don't see any start bit
  	// so interrupt is probably not for us
  	if ( !rx_pin_read() ) {
  		// Wait approximately 1/2 of a bit width to "center" the sample
  		tunedDelay(_rx_delay_centering);
  
  		// Read each of the 8 bits
  		for (uint8_t i = 0x1; i; i <<= 1) {
  			tunedDelay(_rx_delay_intrabit);
  			uint8_t noti = ~i;
  			if (rx_pin_read())
  				d |= i;
  			else // else clause added to ensure function timing is ~balanced
  				d &= noti;
  		}
  
  		// skip the stop bit
  		tunedDelay(_rx_delay_stopbit);
  
  		// if buffer full, set the overflow flag and return
  		if (((_receive_buffer_tail + 1) & _SS_RX_BUFF_MASK) != _receive_buffer_head) {  // circular buffer
  			// save new data in buffer: tail points to where byte goes
  			_receive_buffer[_receive_buffer_tail] = d; // save new byte
  			_receive_buffer_tail = (_receive_buffer_tail + 1) & _SS_RX_BUFF_MASK;  // circular buffer
  		} else {
  			_buffer_overflow = true;
  		}
  	}
  }
#endif



//
// Public methods
//

void ssBegin(long speed) {
	unsigned i;

	_receive_buffer_head = _receive_buffer_tail = 0;
	_buffer_overflow = false;
	SERDDR |= (1<<TXPIN); // set TX for output
	SERDDR &= ~(1<<RXPIN); // set RX for input
	SERPORT |= (1<<TXPIN)|(1<<RXPIN); // assumes no inverse logic

	for (i = 0; i < sizeof(table) / sizeof(table[0]); ++i) {
		long baud = pgm_read_dword(&table[i].baud);
		if (baud == speed) {
			_rx_delay_centering = pgm_read_word(&table[i].rx_delay_centering);
			_rx_delay_intrabit = pgm_read_word(&table[i].rx_delay_intrabit);
			_rx_delay_stopbit = pgm_read_word(&table[i].rx_delay_stopbit);
			_tx_delay = pgm_read_word(&table[i].tx_delay);
			#ifndef TX_ONLY
  			// Set up RX interrupts, but only if we have a valid RX baud rate
  			GIMSK |= (1<<PCIE);
  			PCMSK |= (1<<RXPIN);
  			tunedDelay(_tx_delay);
  			sei();
      #endif
			return;
		}
	}

	// No valid rate found
	// Indicate an error
}

#ifndef TX_ONLY
  void ssEnd() {
  	PCMSK = 0;
  }

// Read data from buffer
  int ssRead() {
  	// Empty buffer?
  	if (_receive_buffer_head == _receive_buffer_tail)
  		return -1;
  
  	// Read from "head"
  	uint8_t d = _receive_buffer[_receive_buffer_head]; // grab next byte
  	_receive_buffer_head = (_receive_buffer_head + 1) & _SS_RX_BUFF_MASK; // circular buffer
  	return d;
  }

  int ssAvailable() {
  	return (_receive_buffer_tail + _SS_MAX_RX_BUFF - _receive_buffer_head) & _SS_RX_BUFF_MASK; // circular buffer
  }
  
  bool ssOverflow(void) {
  	bool ret = _buffer_overflow;
  	_buffer_overflow = false;
  	return ret;
  }
#endif

size_t ssWrite(uint8_t b) {
	if (_tx_delay == 0) {
		//setWriteError();
		return 0;
	}

	uint8_t oldSREG = SREG; // store interrupt flag
	cli();	// turn off interrupts for a clean txmit

	// Write the start bit
	SERPORT &= ~(1<<TXPIN); // tx pin low
	tunedDelay(_tx_delay + XMIT_START_ADJUSTMENT);

	// Write each of the 8 bits
	for (byte mask = 0x01; mask; mask <<= 1) {
		if (b & mask) // choose bit
			SERPORT |= (1<<TXPIN); // tx pin high, send 1
		else
			SERPORT &= ~(1<<TXPIN); // tx pin low, send 0

		tunedDelay(_tx_delay);
	}
	SERPORT |= (1<<TXPIN); // tx pin high, restore pin to natural state

	//sei();
	SREG = oldSREG; // turn interrupts back on
	tunedDelay(_tx_delay);

	return 1;
}

#ifndef TX_ONLY
  void ssFlush() {
    uint8_t oldSREG = SREG; // store interrupt flag
    cli();
    _receive_buffer_head = _receive_buffer_tail = 0;
    SREG = oldSREG; // restore interrupt flag
    //sei();
  }
  
  int ssPeek() {
    // Empty buffer?
    if (_receive_buffer_head == _receive_buffer_tail)
      return -1;
  
    // Read from "head"
    return _receive_buffer[_receive_buffer_head];
  }
#endif

// ------------------------------------------------------------------------------
// additional Functions
// ------------------------------------------------------------------------------

void ssPrint(const char* str) {
    while (*str) {
        ssWrite(*str++);
    }
}

void ssPrint(const __FlashStringHelper *ifsh) {
    const char *p = (const char *)ifsh;
    while (1) {
        unsigned char c = pgm_read_byte(p++);
        if (c == 0) break;
        ssWrite(c);
    }
}

void ssPrint(int num) {
    char buffer[10];
    itoa(num, buffer, 10);
    ssPrint(buffer);
}

void ssPrintHex(int num) {
    char buf[5]; // 4 Ziffern + Nullterminierung
    snprintf(buf, sizeof(buf), "%04X", num);
    ssPrint(buf);
}

void ssPrintln(const char* str) {
    ssPrint(str);
    ssWrite('\r');
    ssWrite('\n');
}

void ssPrintln(const __FlashStringHelper *ifsh) {
    ssPrint(ifsh);
    ssWrite('\r');
    ssWrite('\n');
}

void ssPrintln(int num) {
    ssPrint(num);
    ssWrite('\r');
    ssWrite('\n');
}

void ssPrintlnHex(int num) {
    ssPrintHex(num);
    ssPrint("\r\n");
}

void ssPrintln() {
    ssWrite('\r');
    ssWrite('\n');
}
